/*
 * This program triggers an #BP exception and verifies that the instruction
 * pointer in the trap frame is after the "int3" instruction.
 *
 * In other words it makes sure that the #BP exception is a "trap-style"
 * exception.
 *
 * More debug information will be available on the console by setting
 * 'sysctl machdep.uprintf_signal' to '1'.
 */
#include <sys/types.h>
#include <sys/sysctl.h>

#include <x86/psl.h>
#include <machine/cpufunc.h>

#include <stdio.h>
#include <signal.h>
#include <assert.h>

void my_int3(void);
void before_int3(void) __asm(__STRING(before_int3));
void after_int3(void) __asm(__STRING(after_int3));

static int got_breakpoint;
static int breakpoint_code;
static void *breakpoint_addr;

static void
breakpoint_handler(int num, siginfo_t *info, void *uap)
{
	got_breakpoint = 1;
	breakpoint_code = info->si_code;
	breakpoint_addr = info->si_addr;
}

int
main(void)
{
	u_long flags;
	int error;
	struct sigaction sa;

	sa.sa_sigaction = breakpoint_handler;
	sa.sa_flags = SA_SIGINFO;
	sigemptyset(&sa.sa_mask);
	error = sigaction(SIGTRAP, &sa, NULL);
	assert(error == 0);

	my_int3();

	printf("before_int3: %p\n", before_int3);
	printf("after_int3: %p\n", after_int3);
	printf("breakpoint_addr: %p\n", breakpoint_addr);
	printf("breakpoint_code = %d\n", breakpoint_code);
	printf("got_breakpoint: %d\n", got_breakpoint);
	printf("\n");

	if (got_breakpoint && breakpoint_code == TRAP_BRKPT &&
	    breakpoint_addr == after_int3) {
		printf("success\n");
	} else {
		printf("failed\n");
	}
}
